use std::io::Read;

use craftflow_nbt::DynNBT;
use rust_mc_proto::{DataReader, DataWriter, Packet};
use serde::{Deserialize, Serialize};

use super::ServerError;

pub mod component;
pub mod slot;
pub mod sound;

// Трейт для чтения NBT-совместимых приколов
pub trait RWNBT<'a, T: Serialize + Deserialize<'a>>: DataReader + DataWriter {
	fn read_nbt(&mut self) -> Result<T, ServerError>;
	fn write_nbt(&mut self, val: &T) -> Result<(), ServerError>;
}

impl<'a> RWNBT<'a, DynNBT> for Packet {
	fn read_nbt(&mut self) -> Result<DynNBT, ServerError> {
		let mut data = Vec::new();
		let pos = self.get_ref().position();
		self
			.get_mut()
			.read_to_end(&mut data)
			.map_err(|_| ServerError::DeNbt)?;
		let (remaining, value) = craftflow_nbt::from_slice(&data).map_err(|_| ServerError::DeNbt)?;
		self
			.get_mut()
			.set_position(pos + (data.len() - remaining.len()) as u64);
		Ok(value)
	}

	fn write_nbt(&mut self, val: &DynNBT) -> Result<(), ServerError> {
		craftflow_nbt::to_writer(self.get_mut(), val).map_err(|_| ServerError::SerNbt)?;
		Ok(())
	}
}

pub trait RWPosition: DataReader + DataWriter {
	fn read_position(&mut self) -> Result<(i64, i64, i64), ServerError>;
	fn write_position(&mut self, x: i64, y: i64, z: i64) -> Result<(), ServerError>;
}

impl<T: DataReader + DataWriter> RWPosition for T {
	fn read_position(&mut self) -> Result<(i64, i64, i64), ServerError> {
		let val = self.read_long()?;
		Ok((val >> 38, val << 52 >> 52, val << 26 >> 38))
	}

	fn write_position(&mut self, x: i64, y: i64, z: i64) -> Result<(), ServerError> {
		Ok(self.write_long(((x & 0x3FFFFFF) << 38) | ((z & 0x3FFFFFF) << 12) | (y & 0xFFF))?)
	}
}

#[derive(Clone)]
pub enum IdOr<T> {
	Id(i32),
	Or(T),
}

#[derive(Clone)]
pub enum IdSet {
	Tag(String),
	Ids(Vec<u32>),
}

pub trait RWIdSet: DataReader + DataWriter {
	fn read_id_set(&mut self) -> Result<IdSet, ServerError>;
	fn write_id_set(&mut self, val: IdSet) -> Result<(), ServerError>;
}

impl<T: DataReader + DataWriter> RWIdSet for T {
	fn read_id_set(&mut self) -> Result<IdSet, ServerError> {
		let ty = self.read_varint()?;
		if ty == 0 {
			Ok(IdSet::Tag(self.read_string()?))
		} else {
			let len = ty - 1;
			let mut ids = Vec::new();

			for _ in 0..len {
				ids.push(self.read_u32_varint()?);
			}

			Ok(IdSet::Ids(ids))
		}
	}

	fn write_id_set(&mut self, val: IdSet) -> Result<(), ServerError> {
		match val {
			IdSet::Tag(tag) => {
				self.write_varint(0)?;
				self.write_string(&tag)?;
			}
			IdSet::Ids(ids) => {
				self.write_usize_varint(ids.len() + 1)?;
				for id in ids {
					self.write_u32_varint(id)?;
				}
			}
		}
		Ok(())
	}
}

#[derive(Clone)]
pub struct Property {
	pub name: String,
	pub value: String,
	pub signature: Option<String>,
}

#[derive(Clone, PartialEq, Eq, PartialOrd, Ord)]
pub enum DyeColor {
	White,
	Orange,
	Magenta,
	LightBlue,
	Yellow,
	Lime,
	Pink,
	Gray,
	LightGray,
	Cyan,
	Purple,
	Blue,
	Brown,
	Green,
	Red,
	Black,
}

pub trait RWDyeColor: DataReader + DataWriter {
	fn read_dye_color(&mut self) -> Result<DyeColor, ServerError>;
	fn write_dye_color(&mut self, val: DyeColor) -> Result<(), ServerError>;
}

impl<T: DataReader + DataWriter> RWDyeColor for T {
	fn read_dye_color(&mut self) -> Result<DyeColor, ServerError> {
		match self.read_usize_varint()? {
			0 => Ok(DyeColor::White),
			1 => Ok(DyeColor::Orange),
			2 => Ok(DyeColor::Magenta),
			3 => Ok(DyeColor::LightBlue),
			4 => Ok(DyeColor::Yellow),
			5 => Ok(DyeColor::Lime),
			6 => Ok(DyeColor::Pink),
			7 => Ok(DyeColor::Gray),
			8 => Ok(DyeColor::LightGray),
			9 => Ok(DyeColor::Cyan),
			10 => Ok(DyeColor::Purple),
			11 => Ok(DyeColor::Blue),
			12 => Ok(DyeColor::Brown),
			13 => Ok(DyeColor::Green),
			14 => Ok(DyeColor::Red),
			15 => Ok(DyeColor::Black),
			_ => Err(ServerError::WrongPacket),
		}
	}
	fn write_dye_color(&mut self, val: DyeColor) -> Result<(), ServerError> {
		self.write_usize_varint(val as usize)?;
		Ok(())
	}
}
